"use strict"

class Bittrex extends Exchange {
	static meta = {
		aliases: [
			"BITTREX",
		],
		endpoint: "https://bittrex.com/api/",
		fields: {
			public: {
				label: "Key",
			},
			private: {
				label: "Secret",
			},
		},
		name: "Bittrex",
		patterns: [],
		permissions: {
			origins: [
				"https://bittrex.com/api/*"
			],
		},
		subscriptions: {
			active: [
				"e3fdd77de3a682bfcfd8dc3000a25a5b",
				"4db6261b067dabf8ffb0924c18d95d06",
			],
			inactive: [
				"bitre1",
				"bitre2",
			],
		},
		version: "v1.1",
		fee: 0.0025,
		website: "https://bittrex.com/",
		desc: "Bittrex is a global leader in the blockchain revolution. They operate the premier U.S.-based blockchain trading platform.",
		testSymbol: "BTCUSDT"
	}
	constructor() {
		super(Bittrex.meta)
	}

	*get(resource, params = {}) {
		resource = this.meta.endpoint + this.meta.version + resource

		const headers = {}
		if (!resource.includes("/public/")) {
			const credentials = this.getCredentials("private")
			params.apikey = credentials.public
			params.nonce = this.getNonce()

			let sha = new jsSHA("SHA-512", "TEXT")
			sha.setHMACKey(credentials.private, "TEXT")
			sha.update(resource + '?' + serialize(params))
			headers["apisign"] = sha.getHMAC("HEX")
		} else this.hasAccess()

		const response = yield request.bind(this, 'GET', resource, params, headers)
		if (!response || !response.success || response.message) {
			throw new Error(response.message || "Unknown connection error with API endpoint!")
		}
		return response.result
	}

	*account(currency) {
		const data = yield* this.get("/account/getbalances")
		let balances = {}
		data.forEach(item => {
			if (item.Balance || item.Currency == currency) {
				balances[item.Currency] = item
			}
		})
		balances[currency] = balances[currency] || {Balance: 0, Available: 0}
		return balances
	}

	*symbolInfo() {
		const symbolCache = this.getSymbolCache()
		if (symbolCache._refresh) {
			const response = yield* this.get("/public/getmarkets")

			response.forEach(info => info.IsActive && (symbolCache[info.MarketName] = info))
			let baseCurrencies = new Set(response.map(info => info.BaseCurrency))
			symbolCache.currencies = [...baseCurrencies].join('|')
			this.updatedSymbolCache()
		}
		return symbolCache
	}

	*symbolPair(symbol) {
		symbol = symbol.toUpperCase()
		const market = yield* this.symbolInfo()
		const currencies = market.currencies
		let main = null
		let pair = null

		let reg = new RegExp('^(' + currencies + ')-?(.*?)$', 'i')
		let result = reg.exec(symbol)
		if (result) {
			main = result[1]
			pair = result[2]
		}
		reg = new RegExp('^(.*?)-?(' + currencies + ')$', 'i')
		result = reg.exec(symbol)
		if (result) {
			main = result[2]
			pair = result[1]
		}
		if (!main || !pair) {
			throw new Error(`Unknown market symbol: ${symbol} – Use the 'Symbols' button in the exchange settings or alert editor to browse for valid symbols!`)
		}

		return {
			main: main,
			pair: pair,
			precision: main === "USD" ? 3 : 8,
			symbol: main + '-' + pair
		}
	}

	*symbolTicker(symbol) {
		const ticker = yield* this.get("/public/getticker", {market: symbol})
		return ticker && {
			ask: Number(ticker.Ask),
			bid: Number(ticker.Bid),
			mid: (Number(ticker.Ask) + Number(ticker.Bid)) / 2,
			last: Number(ticker.Last)
		}
	}

	*trade(cmd, ohlc = {}) {
		if (cmd.isMargin) {
			throw new Error(this.getName() + " does not support Margin trading!")
		}
		const market = yield* this.symbolPair(cmd.s)
		const ticker = (cmd.pc && !cmd.up && this.getPrice(market.symbol)) || (yield* this.symbolTicker(market.symbol))
		if (!ticker) {
			throw new ReferenceError(`Ticker ${market.symbol} is not available!`)
		}
		this.checkSlip(cmd, ohlc, market.symbol, ticker)
		if (cmd.up) {
			this.updatePrice(market.symbol, ticker)
			if (!cmd.b && !cmd.ub) {
				return ticker
			}
		}

		const currency = cmd.isBid ? market.main : market.pair
		let balance = {available: 0, balance: 0}
		if (cmd.q.wasPercent() || cmd.ub) {
			balance = cmd.bc && !cmd.ub && this.getBalance(currency)
			const balances = !balance && (yield* this.account(currency))
			balance = balance || balances[currency]
			if (!balance) {
				throw new ReferenceError(`Account Balance ${currency} not available!`)
			}
			if (cmd.minb && balance.Available < cmd.minb._().reference(balance.Balance).resolve()) {
				throw new Error("Not enough balance available! (minb was specified)")
			} else if (cmd.maxb && balance.Available > cmd.maxb._().reference(balance.Balance).resolve()) {
				throw new Error("Too much balance available! (maxb was specified)")
			}
			if (!cmd.bc || cmd.ub) {
				this.updateBalance(balances, currency)
				if (cmd.ub) {
					return balance
				}
			}
		}

		const first = ticker[cmd.pr] || ohlc[cmd.pr] || ((cmd.isBid && cmd.t !== "market") || (cmd.isAsk && cmd.t === "market") ? ticker.bid : ticker.ask)
		let price = cmd.p._().relative(first).minmax(cmd.minp, cmd.maxp).resolve(market.precision)
		if (cmd.fp) {
			price = cmd.fp.resolve(market.precision)
		}
		if (price <= 0) {
			this.warn("your p parameter might be incorrect! (price of 0 calculated)")
		}

		let available = cmd.y === "equity" ? balance.Balance : balance.Available
		available = (1 - this.meta.fee) * available
		if (cmd.isBid) {
			available /= price
		}

		ohlc[cmd.qr] && this.info(`using custom quantity multiplier: ${ohlc[cmd.qr]} (${cmd.qr})`)
		const method = (cmd.isBid ? "buy" : "sell") + (cmd.t === "market" ? "market" : "limit")
		const order = {
			quantity: cmd.q._().div(cmd.u === "currency" && !cmd.q.wasPercent() ? price : 1)
				.reference(available).mul(ohlc[cmd.qr] || 1).minmax(cmd.minq, cmd.maxq).resolve(market.precision),
			market: market.symbol,
			rate: price
		}
		if (cmd.lq && ohlc.left) {
			this.info("using leftover quantity from last order: "+ohlc.left)
			order.quantity = ohlc.left
		}
		if (order.quantity <= 0) {
			this.warn("your q parameter might be incorrect! (quantity of 0 calculated)")
		}

		this.info(`placing ${method} order: `+stringify(order))
		if (cmd.d) {
			this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
			return order
		}
		const result = yield* this.get("/market/" + method, order)
		cmd.id && result && result.id && this.cacheId(result.id, cmd.id)
		return result
	}

	*ordersCancel(cmd) {
		const market = yield* this.symbolPair(cmd.s)
		let orders = yield* this.get("/market/getopenorders", {market: market.symbol})
		if (!orders || !orders.filter) {
			this.info("unable to get order info!")
			return false
		} else if (!orders.length) {
			this.info("no orders open!")
			return false
		}

		const ids = (cmd.id || "").split(',')
		orders = orders.filter(order => {
			if ((cmd.isBid && !order.OrderType.includes("BUY")) || (cmd.isAsk && !order.OrderType.includes("SELL"))) {
				return false
			}
			if (cmd.fp && cmd.fp.compare(order.Limit)) {
				return false
			}
			if (cmd.id && !this.checkId(order.id, ids)) {
				return false
			}
			return true
		})

		if (cmd.cm._().reference(orders.length).getMax() < orders.length) {
			const sort = {
				"newest": 	["Opened", true],
				"oldest": 	["Opened", false],
				"highest": 	["Limit", true],
				"lowest": 	["Limit", false],
				"biggest": 	["Quantity", true],
				"smallest": ["Quantity", false]
			}
			cmd.cmo === 'random' ? shuffle(orders) : sortByIndex(orders, sort[cmd.cmo][0], sort[cmd.cmo][1])
			orders = orders.slice(0, cmd.cm.resolve(0))
		}
		if(!orders.length) {
			this.info("open orders don't match any criteria!")
			return []
		}

		this.info((cmd.ch ? "orders found: " : "canceling orders: ")+stringify(orders))
		if (cmd.ch || cmd.d) {
			!cmd.ch && this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
			return orders
		}

		let cancelOrders = []
		for (const order of orders) {
			const result = yield* this.get("/market/cancel", {uuid: order.OrderUuid})
			if (result && result.success) {
				cancelOrders.push(order)
				this.removeId(order.OrderUuid)
			}
		}
		return cancelOrders
	}

	*positionsCloseAll() {
		throw new ReferenceError(this.getName() + " does not support Margin trading!")
	}
}

Broker.add(new Bittrex())
